# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TrackingEvent',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', models.CharField(max_length=255)),
                ('tracking_id', models.CharField(max_length=255)),
                ('client_id', models.CharField(max_length=255)),
                ('user_agent', models.TextField()),
                ('data', models.TextField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('captured_at', models.DateTimeField(null=True)),
                ('relayed_at', models.DateTimeField(null=True)),
                ('registered_at', models.DateTimeField(null=True)),
                ('status', models.CharField(max_length=255, null=True, choices=[(b'captured', b'Captured'), (b'relayed', b'Relayed'), (b'registered', b'Registered')])),
            ],
            options={
                'ordering': ['created_at'],
            },
        ),
    ]



from ..settings.base import *


# Disable debug mode

DEBUG = False
TEMPLATE_DEBUG = False

ROOT_URLCONF = 'garelay.tracker.urls'

CELERYBEAT_SCHEDULE = {
    'relay-events': {
        'task': 'garelay.tracker.tasks.relay_events',
        'schedule': timedelta(minutes=1),
    },
}

try:
    from .local import *
except ImportError:
    pass



from __future__ import absolute_import

import os

from celery import Celery

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'garelay.settings.production')

from django.conf import settings

app = Celery('proj')

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object('django.conf:settings')
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)



import json

from django.test import TestCase
from django.test.client import Client
from django.utils import timezone
from django.core.urlresolvers import reverse
from django.core.serializers.json import DjangoJSONEncoder

from garelay.tracker.models import TrackingEvent


class ServerTest(TestCase):

    def setUp(self):
        self.client = Client()

    def test_tracking_event(self):
        event = TrackingEvent.objects.create(uuid='the-uuid')
        event.update_fields({
            'tracking_id': 'tracking_id',
            'client_id': 'client_id',
            'user_agent': 'user_agent',
            'data': {
                'foo': 'bar'
            },
            'created_at': timezone.now(),
            'updated_at': timezone.now(),
            'captured_at': timezone.now(),
            'relayed_at': None,
            'registered_at': None,
        })
        event.save()
        self.assertFalse(event.relayed_at)
        self.assertFalse(event.registered_at)
        self.assertFalse(event.status)
        response = self.client.post(
            reverse('server'),
            data=json.dumps([event.to_dict()], cls=DjangoJSONEncoder),
            content_type='application/json')
        data = json.loads(response.content)
        self.assertTrue(data, [{'uuid': 'the-uuid'}])
        reloaded_event = TrackingEvent.objects.get(pk=event.pk)
        self.assertTrue(reloaded_event.relayed_at)
        self.assertFalse(reloaded_event.registered_at)
        self.assertEqual(reloaded_event.status, 'captured')



from .base import *


# Disable debug mode

DEBUG = False
TEMPLATE_DEBUG = False


# Compress static files offline
# http://django-compressor.readthedocs.org/en/latest/settings/#django.conf.settings.COMPRESS_OFFLINE

COMPRESS_OFFLINE = True


# Send notification emails as a background task using Celery,
# to prevent this from blocking web server threads
# (requires the django-celery package):
# http://celery.readthedocs.org/en/latest/configuration.html

# import djcelery
#
# djcelery.setup_loader()
#
# CELERY_SEND_TASK_ERROR_EMAILS = True
# BROKER_URL = 'redis://'

try:
    from .local import *
except ImportError:
    pass



import json

import responses

from django.conf import settings
from django.test import TestCase
from django.utils import timezone

from ..models import TrackingEvent
from ..tasks import relay_events


class TaskTest(TestCase):

    def setUp(self):
        responses.add(responses.POST, settings.GARELAY_SERVER,
                      status=200, content_type='application/json',
                      body=json.dumps([
                          {'uuid': 'uuid-1'},
                          {'uuid': 'uuid-3'},
                      ]))

    @responses.activate
    def test_register_events(self):
        TrackingEvent.objects.create(
            uuid='uuid-1',
            status='captured', data=json.dumps({}),
            captured_at=timezone.now())
        TrackingEvent.objects.create(
            uuid='uuid-2',
            status='relayed', data=json.dumps({}),
            captured_at=timezone.now())
        TrackingEvent.objects.create(
            uuid='uuid-3',
            status='captured', data=json.dumps({}),
            captured_at=timezone.now())
        self.assertEqual(
            TrackingEvent.objects.filter(status='relayed').count(), 1)
        relay_events()
        self.assertEqual(
            TrackingEvent.objects.filter(status='relayed').count(), 3)



import os

from django.conf.urls import patterns, include, url
from django.conf.urls.static import static
from django.conf import settings
from django.contrib import admin


urlpatterns = patterns(
    '',
    url(r'^admin/', include(admin.site.urls)),
)

if not os.environ.get('GARELAY_NO_TRACKER'):
    urlpatterns += patterns(
        '',
        url(r'^', include('garelay.tracker.urls')),
    )

if not os.environ.get('GARELAY_NO_SERVER'):
    urlpatterns += patterns(
        '',
        url(r'^', include('garelay.server.urls')),
    )

if settings.DEBUG:
    from django.contrib.staticfiles.urls import staticfiles_urlpatterns

    urlpatterns += staticfiles_urlpatterns()
    urlpatterns += static(
        settings.MEDIA_URL + 'images/',
        document_root=os.path.join(settings.MEDIA_ROOT, 'images'))



from django.test import TestCase
from django.utils import timezone
from datetime import timedelta
from ..models import TrackingEvent

from mock import patch

from UniversalAnalytics import Tracker


class ModelTest(TestCase):

    def update_event_fields(self, event, **fields):
        defaults = {
            'tracking_id': 'tracking_id',
            'client_id': 'client_id',
            'user_agent': 'user_agent',
            'data': {
                'dt': 'Document Title'
            },
            'created_at': timezone.now(),
            'updated_at': timezone.now(),
            'captured_at': timezone.now(),
            'relayed_at': timezone.now(),
            'registered_at': timezone.now(),
        }
        defaults.update(fields)
        event.update_fields(defaults)
        return defaults

    def test_update_fields(self):
        event = TrackingEvent.objects.create(uuid='the-uuid')
        self.update_event_fields(event, id=-1, status='captured')
        self.assertTrue(event.tracking_id)
        self.assertTrue(event.client_id)
        self.assertTrue(event.user_agent)
        self.assertTrue(event.data)
        self.assertTrue(event.created_at)
        self.assertTrue(event.updated_at)
        self.assertTrue(event.captured_at)
        self.assertTrue(event.relayed_at)
        self.assertTrue(event.registered_at)
        self.assertNotEqual(event.status, 'captured')
        self.assertNotEqual(event.id, -1)

    def test_to_dict(self):
        event = TrackingEvent.objects.create(uuid='the-uuid')
        fields = self.update_event_fields(event)
        self.assertNotEqual(
            event.to_dict(),
            fields)

    @patch.object(Tracker.Tracker, 'send')
    def test_register(self, mocked_send):
        queue_time = timezone.now() - timedelta(minutes=10)
        event = TrackingEvent.objects.create()
        self.update_event_fields(event, captured_at=queue_time)
        event.save()
        event.register()
        [call] = mocked_send.call_args_list
        [args, kwargs] = call
        self.assertEqual(kwargs, {})
        hittype, data = args
        self.assertEqual(hittype, 'pageview')
        self.assertEqual(data['dt'], 'Document Title')
        # Allowing for some minimal CPU time
        self.assertTrue(10 * 60 * 1000 < int(data['qt']) < 10.1 * 60 * 1000)



import json

from django.test import TestCase
from django.test.client import Client
from django.core.urlresolvers import reverse

from ..models import TrackingEvent


class ViewsTest(TestCase):

    def setUp(self):
        self.client = Client(
            HTTP_USER_AGENT='The Agent',
            HTTP_ACCEPT_LANGUAGE='en-gb',
            HTTP_REFERER='/foo')

    def test_tracker_response_headers(self):
        response = self.client.get(reverse('tracker', kwargs={
            'tracking_id': 'UA-F00-1',
            'path': '',
        }))
        self.assertEqual(response['Content-Type'], 'image/gif')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Cache-Control'], 'max-age=0')
        self.assertTrue(response['Expires'])

    def test_tracking_event(self):
        self.assertEqual(TrackingEvent.objects.count(), 0)
        path = reverse('tracker', kwargs={
            'tracking_id': 'UA-F00-1',
            'path': '',
        })
        self.client.get(path)
        [event] = TrackingEvent.objects.all()
        self.assertTrue(event.uuid)
        self.assertTrue(event.created_at)
        self.assertTrue(event.updated_at)
        self.assertTrue(event.captured_at)
        self.assertFalse(event.relayed_at)
        self.assertFalse(event.registered_at)
        self.assertEqual(event.tracking_id, 'UA-F00-1')
        self.assertEqual(event.client_id, self.client.session['tracker_uuid'])
        self.assertEqual(event.user_agent, 'The Agent')
        self.assertEqual(event.status, 'captured')
        self.assertEqual(json.loads(event.data), {
            'ul': 'en-gb',
            'dr': '/foo',
            'uip': '127.0.0.1',
            'dp': '',
        })
        self.assertEqual(TrackingEvent.objects.count(), 1)

    def test_tracking_path(self):
        path = reverse('tracker', kwargs={
            'tracking_id': 'UA-F00-1',
            'path': '/foo/bar/baz.html',
        })
        self.client.get(path)
        [event] = TrackingEvent.objects.all()
        self.assertEqual(
            json.loads(event.data)['dp'],
            '/foo/bar/baz.html')



from StringIO import StringIO

from django.test import TestCase
from django.conf import settings

from garelay.management.commands import relay_events, register_events


class CommandsTest(TestCase):

    def test_relay_events(self):
        cmd = relay_events.Command()
        cmd.stdout = StringIO()
        cmd.handle()
        self.assertEqual(
            cmd.stdout.getvalue(),
            'Relaying 0 events in batches of %s.\nDone.\n' % (
                settings.GARELAY_RELAY_BATCH_SIZE,))

    def test_register_events(self):
        cmd = register_events.Command()
        cmd.stdout = StringIO()
        cmd.handle()
        self.assertEqual(
            cmd.stdout.getvalue(),
            'Registering 0 events in batches of %s.\nDone.\n' % (
                settings.GARELAY_REGISTER_BATCH_SIZE,))



"""
Django settings for base garelay.

For more information on this file, see
https://docs.djangoproject.com/en/1.7/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.7/ref/settings/
"""

from datetime import timedelta
from os.path import abspath, dirname, join
from django.conf import global_settings
from django.utils.translation import ugettext_lazy as _

# Absolute filesystem path to the Django project directory:
PROJECT_ROOT = dirname(dirname(dirname(abspath(__file__))))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.7/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "0o9t7z)z6i=bpzfz5!u3pp45r_d2t7-x99pw#omqzi&ywbze30"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = ['*']

# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'garelay',
    'garelay.tracker',
    'garelay.server',

    'raven.contrib.django.raven_compat',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'garelay.urls'
WSGI_APPLICATION = 'garelay.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.7/ref/settings/#databases

# SQLite (simplest install)
import dj_database_url
DATABASES = {'default': dj_database_url.config(
    default='sqlite:///%s' % (join(PROJECT_ROOT, 'db.sqlite3'),))}

# PostgreSQL (Recommended, but requires the psycopg2 library and Postgresql
#             development headers)
# DATABASES = {
#     'default': {
#         'ENGINE': 'django.db.backends.postgresql_psycopg2',
#         'NAME': 'base',
#         'USER': 'postgres',
#         'PASSWORD': '',
#         'HOST': '',  # Set to empty string for localhost.
#         'PORT': '',  # Set to empty string for default.
#         # number of seconds database connections should persist for
#         'CONN_MAX_AGE': 600,
#     }
# }

# CELERY stuff
BROKER_URL = 'redis://localhost:6379/0'
CELERY_RESULT_BACKEND = 'redis://localhost:6379/0'
CELERYBEAT_SCHEDULE = {
    'relay-events': {
        'task': 'garelay.tracker.tasks.relay_events',
        'schedule': timedelta(minutes=1),
    },
    'register-events': {
        'task': 'garelay.server.tasks.register_events',
        'schedule': timedelta(minutes=1),
    }
}

CELERY_ACCEPT_CONTENT = ['json']
CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'

# GARELAY stuff
GARELAY_SERVER = 'http://www.example.org/server'
GARELAY_CREDENTIALS = ('username', 'password')
GARELAY_RELAY_BATCH_SIZE = 100
GARELAY_REGISTER_BATCH_SIZE = 1000
GARELAY_RELAY_TIMEOUT = 10

# Internationalization
# https://docs.djangoproject.com/en/1.7/topics/i18n/

LANGUAGE_CODE = 'en-gb'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True

# Native South African languages are currently not included in the default
# list of languges in django
# https://github.com/django/django/blob/master/django/conf/global_settings.py#L50
LANGUAGES = global_settings.LANGUAGES + (
    ('zu', _('Zulu')),
    ('xh', _('Xhosa')),
    ('st', _('Sotho')),
    ('ve', _('Venda')),
    ('tn', _('Tswana')),
    ('ts', _('Tsonga')),
    ('ss', _('Swati')),
    ('nr', _('Ndebele')),
)

LOCALE_PATHS = (
    join(PROJECT_ROOT, "locale"),
)

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.7/howto/static-files/

STATIC_ROOT = join(PROJECT_ROOT, 'static')
STATIC_URL = '/static/'

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'compressor.finders.CompressorFinder',
)

MEDIA_ROOT = join(PROJECT_ROOT, 'media')
MEDIA_URL = '/media/'


# Django compressor settings
# http://django-compressor.readthedocs.org/en/latest/settings/

COMPRESS_PRECOMPILERS = (
    ('text/x-scss', 'django_libsass.SassCompiler'),
)


# Template configuration

TEMPLATE_CONTEXT_PROCESSORS = global_settings.TEMPLATE_CONTEXT_PROCESSORS + (
    'django.core.context_processors.request',
    'molo.core.context_processors.locale',
)



from django.conf import settings

from garelay import celery_app
from garelay.tracker.models import TrackingEvent


@celery_app.task(ignore_result=True)
def register_events():
    # Limit to 1000 requests registered at GA per minute.
    # That's roughly 16 per scond
    events = TrackingEvent.objects.all().exclude(status='registered')
    for event in events[:settings.GARELAY_REGISTER_BATCH_SIZE]:
        event.register()
        event.save()



from django.conf.urls import patterns, url

from .views import tracker


urlpatterns = patterns(
    '',
    url(r'^(?P<path>.*)/tracker-(?P<tracking_id>[A-Za-z0-9\-]+)\.gif',
        tracker, name='tracker'),
)



from django.core.management.base import BaseCommand
from django.conf import settings

from garelay.server import tasks
from garelay.tracker.models import TrackingEvent


class Command(BaseCommand):
    help = 'Register events with Google Analytics.'

    def handle(self, *args, **kwargs):
        events = TrackingEvent.objects.all().exclude(status='registered')
        self.stdout.write('Registering %s events in batches of %s.\n' % (
            events.count(),
            settings.GARELAY_REGISTER_BATCH_SIZE))
        tasks.register_events()
        self.stdout.write('Done.\n')



import json

from django.http import HttpResponseBadRequest, JsonResponse
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt
from garelay.tracker.models import TrackingEvent


@csrf_exempt
def server(request):
    try:
        tracking_events = json.loads(request.body)
    except ValueError:
        return HttpResponseBadRequest()

    response = []
    for tracking_event in tracking_events:
        record, _ = TrackingEvent.objects.get_or_create(
            uuid=tracking_event['uuid'])
        record.update_fields(tracking_event)
        record.status = 'captured'
        record.relayed_at = timezone.now()
        record.save()
        response.append({
            'uuid': record.uuid,
        })
    return JsonResponse(response, safe=False)



import json

from django.conf import settings
from django.utils import timezone
from django.core.serializers.json import DjangoJSONEncoder

from garelay import celery_app
from garelay.tracker.models import TrackingEvent

import requests


@celery_app.task(ignore_result=True)
def relay_events():
    pending_events = TrackingEvent.objects.filter(
        status='captured').exclude(status='relayed')
    while pending_events.exists():
        next_batch = pending_events[:settings.GARELAY_RELAY_BATCH_SIZE]
        payload = [event.to_dict() for event in next_batch]
        response = requests.post(
            settings.GARELAY_SERVER,
            auth=settings.GARELAY_CREDENTIALS,
            data=json.dumps(payload, cls=DjangoJSONEncoder),
            headers={
                'content-type': 'application/json',
                'accept': 'application/json',
            },
            timeout=settings.GARELAY_RELAY_TIMEOUT)
        # This will kill the task if we don't have Internet connectivity
        response.raise_for_status()
        uuids = [event['uuid'] for event in response.json()]
        TrackingEvent.objects.filter(uuid__in=uuids).update(
            status='relayed', relayed_at=timezone.now())



from django.core.management.base import BaseCommand
from django.conf import settings

from garelay.tracker import tasks
from garelay.tracker.models import TrackingEvent


class Command(BaseCommand):
    help = 'Relay events to the configured GARELAY_SERVER.'

    def handle(self, *args, **kwargs):
        events = TrackingEvent.objects.filter(
            status='captured').exclude(status='relayed')
        self.stdout.write('Relaying %s events in batches of %s.\n' % (
            events.count(),
            settings.GARELAY_RELAY_BATCH_SIZE))
        tasks.relay_events()
        self.stdout.write('Done.\n')



import json
import logging

from uuid import uuid4

from django.db import models
from django.db.models import signals
from django.utils import timezone
from django.dispatch import receiver

from UniversalAnalytics import Tracker


class TrackingEvent(models.Model):
    uuid = models.CharField(max_length=255)
    tracking_id = models.CharField(max_length=255)
    client_id = models.CharField(max_length=255)
    user_agent = models.TextField()
    data = models.TextField(default='{}')
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    captured_at = models.DateTimeField(null=True)
    relayed_at = models.DateTimeField(null=True)
    registered_at = models.DateTimeField(null=True)
    status = models.CharField(max_length=255, choices=(
        ('captured', 'Captured'),
        ('relayed', 'Relayed'),
        ('registered', 'Registered')
    ), null=True)

    class Meta:
        ordering = ['created_at']

    def update_fields(self, data):
        self.tracking_id = data['tracking_id']
        self.client_id = data['client_id']
        self.user_agent = data['user_agent']
        self.data = json.dumps(data['data'])
        self.captured_at = data['captured_at']
        self.relayed_at = data['relayed_at']
        self.registered_at = data['registered_at']

    def to_dict(self):
        payload = dict(
            (field.name, getattr(self, field.name))
            for field in self._meta.fields
            if field.name not in ['id', 'status'])
        payload['data'] = json.loads(payload['data'])
        return payload

    def register(self):
        tracker = Tracker.create(self.tracking_id,
                                 client_id=self.client_id,
                                 user_agent=self.user_agent)
        payload = json.loads(self.data)

        # back date
        delta = timezone.now() - self.captured_at
        if delta.seconds > (4 * 60 * 60):
            logging.warning('Queue time exceeds 4 hours, '
                            'may be ignored by Google Analytics.')

        # GA requires milliseconds
        payload['qt'] = delta.total_seconds() * 1000

        tracker.send('pageview', payload)
        self.registered_at = timezone.now()
        self.status = 'registered'


@receiver(signals.pre_save, sender=TrackingEvent)
def tracking_event_auto_uuid(sender, instance, **kwargs):
    if not instance.uuid:
        instance.uuid = uuid4().hex



from uuid import uuid4
import json

from django.http import HttpResponse
from django.views.decorators.cache import never_cache
from django.utils import timezone

from .models import TrackingEvent


PIXEL_GIF_DATA = ("R0lGODlhAQABAIAAAAAAAP///yH5"
                  "BAEAAAAALAAAAAABAAEAAAIBRAA7").decode('base64')


def get_tracker_uuid(request):
    return request.session.setdefault('tracker_uuid', uuid4().hex)


@never_cache
def tracker(request, tracking_id, path):
    client_id = get_tracker_uuid(request)
    data = {
        'dp': path,
        'uip': request.META.get('REMOTE_ADDR') or '',
        'dr': request.META.get('HTTP_REFERER') or '',
        'ul': request.META.get('HTTP_ACCEPT_LANGUAGE') or '',
    }
    data.update(request.GET)
    TrackingEvent.objects.create(
        tracking_id=tracking_id,
        client_id=client_id,
        user_agent=request.META.get('HTTP_USER_AGENT') or '',
        data=json.dumps(data),
        captured_at=timezone.now(),
        status='captured')
    return HttpResponse(PIXEL_GIF_DATA, content_type='image/gif')



import json

from django.test import TestCase
from django.utils import timezone

from ..tasks import register_events

from garelay.tracker.models import TrackingEvent

from UniversalAnalytics import Tracker

from mock import patch


class TaskTest(TestCase):

    @patch.object(Tracker.Tracker, 'send')
    def test_register_events(self, mocked_send):
        TrackingEvent.objects.create(
            status='registered', data=json.dumps({}),
            captured_at=timezone.now())
        TrackingEvent.objects.create(
            status='relayed', data=json.dumps({}),
            captured_at=timezone.now())
        register_events()
        self.assertEqual(mocked_send.call_count, 1)
